from os import getcwd
from shutil import rmtree

import click

from .gesetz import Gesetz
from .helpers import analyze as analyze_norm
from .scrapers import *


@click.group()
@click.pass_context
@click.option('-v', '--verbose', count = True, help = 'Enable verbose mode.')
@click.version_option('1.1.0')
def cli(ctx, verbose: int) -> None:
    """
    Utilities for indexing & analyzing german legal norms
    """

    # Ensure context object exists & is dictionary
    ctx.ensure_object(dict)

    # Initialize context object & assign verbose mode
    ctx.obj = {'verbose': verbose}


@cli.command()
@click.pass_context
@click.option('-p', '--provider', multiple = True, help = 'Provider for legal norms.')
@click.option('-d', '--directory', type = click.Path(True), help = 'Target directory.')
@click.option('-w', '--wait', default = 3, help = 'Time between requests (in seconds).')
def build(ctx, provider: tuple, directory: click.Path, wait: int) -> None:
    """
    Builds index
    """

    if ctx.obj['verbose'] > 0:
        click.echo('Selected waiting time: "{}"'.format(wait))

    if not directory:
        if ctx.obj['verbose'] > 0:
            click.echo('No directory selected, using fallback ..')

        directory = getcwd()

    if ctx.obj['verbose'] > 0:
        click.echo('Selected target directory: "{}"'.format(directory))

    databases = [
        {
            'driver': ['buz', 'buzer'],
            'class': Buzer,
            'url': 'buzer.de',
        },
        {
            'driver': ['djo', 'dejure'],
            'class': DejureOnline,
            'url': 'dejure.org',
        },
        {
            'driver': ['gii', 'gesetze'],
            'class': GesetzeImInternet,
            'url': 'gesetze-im-internet.de',
        },
        {
            'driver': ['lex', 'lexparency'],
            'class': Lexparency,
            'url': 'lexparency.de',
        },
    ]

    # Iterate over providers ..
    for db in databases:
        # .. but skip invalid ones
        if provider and not list(set(provider) & set(db['driver'])):
            continue

        if ctx.obj['verbose'] > 0:
            click.echo('Current database: "{}" ..'.format(db['url']))

        driver = db['class'](click.get_app_dir('gesetze'))

        output_file = '{}/{}.json'.format(directory, driver.identifier)

        if ctx.obj['verbose'] > 0:
            click.echo('Saving data in file "{}" ..'.format(output_file))

        driver.scrape(output_file, wait)


@cli.command()
@click.pass_context
def clear(ctx) -> None:
    """
    Clears download cache
    """

    if ctx.obj['verbose'] > 0:
        click.echo('Clearing cache ..', nl = False)

    rmtree(click.get_app_dir('gesetze'))

    if ctx.obj['verbose'] > 0:
        click.echo(' done.')


@cli.command()
@click.pass_context
@click.argument('norm', type = str, nargs = -1)
def analyze(ctx, norm) -> None:
    """
    Analyzes legal NORM
    """

    if ctx.obj['verbose'] > 0:
        click.echo('Analyzing input ..', nl = False)

    result = analyze_norm(' '.join(norm))

    if ctx.obj['verbose'] > 0:
        click.echo(' done.')

    for k, v in result.items():
        click.echo('{}: {}'.format(k.capitalize(), v))



from os.path import dirname, exists, realpath
from abc import ABC, abstractmethod
from typing import Optional

from ..helpers import analyze
from ..utils import load_json


class Driver(ABC):
    """
    Utilities for dealing with providers
    """

    # Individual identifier
    identifier: Optional[str] = None


    def __init__(self, file: Optional[str] = None) -> None:
        """
        Creates 'Driver' instance

        :param file: str Path to (custom) data file

        :raises: Exception Invalid law
        """

        # Set default index file
        if file is None:
            file = '{}/../data/{}.json'.format(dirname(__file__), self.identifier)

        # Fail early if file does not exist
        if not exists(file):
            raise Exception('File does not exist: "{}"'.format(realpath(file)))

        # Load law library data
        self.library = load_json(file)


    def validate(self, string: str) -> bool:
        # Fail early when string is empty
        if not string:
            return False

        # Analyze legal norm
        data = analyze(string)

        # Get lowercase identifier for current law
        identifier = data['gesetz'].lower()

        # Fail early if law is unavailable
        if identifier not in self.library:
            # .. fail early
            return False

        # Return whether current law contains norm
        return data['norm'] in self.library[identifier]['headings']


    def build_title(self, data: dict, mode: str) -> str:
        """
        Builds description for corresponding legal norm (used as `title` attribute)

        :param data: dict Legal data
        :param mode: str Output mode, either 'light', 'normal' or 'full' (default: False)

        :return: str Title attribute
        :raises: Exception Invalid law
        """

        # Get lowercase identifier for current law
        identifier = data['gesetz'].lower()

        # Fail early if law is unavailable
        if identifier not in self.library:
            raise Exception('Invalid law: "{}"'.format(data['gesetz']))

        # Get data about current law
        law = self.library[identifier]

        # Determine `title` attribute
        if mode == 'light':
            return law['law']

        if mode == 'normal':
            return law['title']

        if mode == 'full':
            return law['headings'][data['norm']]

        return ''


    @abstractmethod
    def build_url(self, data: dict) -> str:
        """
        Builds URL for corresponding legal norm (used as `href` attribute)

        :param data: dict Legal data

        :return: str Target URL
        """

        pass



from os.path import exists
import re
import time

from .scraper import Scraper
from ..utils import dump_json


class Buzer(Scraper):
    """
    Utilities for scraping 'buzer.de'
    """

    # Individual identifier
    identifier: str = 'buzer'


    def scrape(self, output_file: str, wait: int = 2) -> None:
        """
        Scrapes 'buzer.de' for legal norms

        :param output_file: str Path to merged data file
        :param wait: int Time to wait before scraping next law

        :return: None
        """

        # Fetch overview page
        html = self.get_html('https://www.buzer.de')

        # Create list of data files
        data_files = []

        # Parse their HTML & iterate over `a` tags ..
        for link in self.bs(html).select('.table1')[0].select('a'):
            # .. extracting data for each law
            law = link.text.strip()

            # Define data file for current law
            data_file = self.md5_file(self.dir, law)

            # If it already exists ..
            if exists(data_file):
                # (1) .. update list of data files
                data_files.append(data_file)

                # (2) .. skip law
                continue

            slug = link['href']

            # .. collecting its information
            node = {
                'law': law,
                'slug': slug,
                'title': '',
                'headings': {},
            }

            # Fetch index page for each law
            law_html = self.get_html(slug)

            # Get title
            title = self.bs(law_html).select('h1')[0].text

            # Strip shorthand from title (if present)
            match = re.match(r'(.*)\s\(.*\)$', title)

            # If heading was found ..
            if match:
                # .. store it as title for current law
                title = match.group(1)

            node['title'] = title

            # Iterate over `li` tags ..
            for heading in self.bs(law_html).find_all('a', attrs={'class': 'preview'}):
                # .. skipping headings without `title` attribute
                if not heading.get('title'):
                    continue

                heading = {
                    'slug': heading['href'][1:],
                    'text': heading.text.replace('§  ', '§ '),
                }

                # Determine section identifier
                match = re.match(r'(?:§+|Art|Artikel)\.?\s*(\d+(?:\w\b)?)', heading['text'], re.IGNORECASE)

                # If section identifier was found ..
                if match:
                    # .. store identifier as key and heading as value
                    node['headings'][match.group(1)] = heading

                # .. otherwise ..
                else:
                    # .. store heading as both key and value
                    node['headings'][heading['text']] = heading

            # Store data record
            # (1) Write data to JSON file
            dump_json(node, data_file)

            # (2) Update list of data files
            data_files.append(data_file)

            # Wait for it ..
            time.sleep(wait)

        # Merge data files
        self.build(data_files, output_file)



from .drivers import *


class Factory:
    """
    Utilities for creating 'Driver' instances
    """

    # List available drivers
    drivers: dict = {
        # (1) 'gesetze-im-internet.de'
        'gesetze': GesetzeImInternet,
        # (2) 'dejure.org'
        'dejure': DejureOnline,
        # (3) 'buzer.de'
        'buzer': Buzer,
        # (4) 'lexparency.de'
        'lexparency': Lexparency,
    }


    def create(self, driver: str) -> Driver:
        """
        Creates a new 'Driver' instance for the given type

        :param driver: str Driver type

        :return: Driver Driver instance
        :raises: Exception Invalid driver type
        """

        # Fail early for invalid drivers
        if driver not in self.drivers:
            raise Exception('Invalid driver type: "{}"'.format(driver))

        # Instantiate object
        return self.drivers[driver]()



from os import getcwd
from os.path import exists
from abc import ABC, abstractmethod
from hashlib import md5
from shutil import rmtree
from typing import Optional

from requests import get
from bs4 import BeautifulSoup

from ..utils import create_path, dump_json, load_json


class Scraper(ABC):
    """
    Utilities for scraping providers
    """

    # Individual identifier
    identifier: Optional[str] = None


    def __init__(self, path: str) -> None:
        """
        Initializes download directory

        :param path: str Path to directory

        :return: None
        """

        # Create data directory
        # (1) Define its path
        self.dir = '{}/.{}'.format(path, self.identifier)

        # (2) Create (if necessary)
        create_path(self.dir)


    def hash(self, file: str) -> str:
        """
        Creates hashed filename

        :param file: str Filename

        :return: str Hash
        """

        return md5(file.encode('utf-8')).hexdigest()


    def get_html(self, url: str) -> str:
        """
        Fetches HTML of given URL

        :param url: str Target URL

        :return: str Target HTML
        """

        file = '{}/{}.html'.format(self.dir, self.hash(url))

        if not exists(file):
            html = get(url).text

            with open(file, 'w') as html_file:
                html_file.write(html)

        else:
            with open(file, 'r') as html_file:
                html = html_file.read()

        return html


    def bs(self, html: str) -> BeautifulSoup:
        """
        Converts HTML to 'BeautifulSoup' object

        :param html: str Target HTML

        :return: bs4.BeautifulSoup 'BeautifulSoup' object
        """

        return BeautifulSoup(html, 'lxml')


    def md5_file(self, path: str, file: str) -> str:
        """
        Determines hashed JSON data file

        :param path: str Data directory
        :param file: str Filename for corresponding law

        :return: str Path to data file
        """

        return '{}/{}.json'.format(path, self.hash(file))


    def build(self, data_files: list, output_file: Optional[str] = None) -> None:
        """
        Merges JSON files & removes them afterwards

        :param data_files: list List of data files
        :param output_file: str Path to merged data file

        :return: None
        """

        # Create data buffer
        data = {}

        # Iterate over data files
        for data_file in data_files:
            # Load data & update data buffer
            node = load_json(data_file)
            data[node['law'].lower()] = node

        # if target file not specified ..
        if output_file is None:
            # .. use current working directory & identifier as fallback
            output_file = '{}/{}.json'.format(getcwd(), self.identifier)

        # Write complete dataset to JSON file
        dump_json(data, output_file)

        # Remove temporary files
        rmtree(self.dir)


    @abstractmethod
    def scrape(self, output_file: str, wait: int = 2) -> None:
        """
        Scrapes website for legal norms

        :param output_file: str Path to merged data file
        :param wait: int Time to wait before scraping next law

        :return: None
        """

        pass



from .driver import Driver
from .buzer import Buzer
from .dejure import DejureOnline
from .gesetze import GesetzeImInternet
from .lexparency import Lexparency


__all__ = [
    # Base class
    'Driver',

    # Driver classes
    'Buzer',
    'DejureOnline',
    'GesetzeImInternet',
    'Lexparency',
]



from typing import Optional, Union


def load_json(json_file: str) -> dict:
    """
    Loads data from JSON file

    :param json_file: str Path to JSON file

    :return: dict Loaded JSON data
    :raises: Exception Error decoding JSON
    """

    from json import load
    from json.decoder import JSONDecodeError

    try:
        with open(json_file, 'r') as file:
            return load(file)

    except JSONDecodeError:
        raise Exception

    return {}


def dump_json(data: Union[dict, list], json_file: str, indent: Optional[int] = None, ensure_ascii: bool = False) -> None:
    """
    Dumps data to JSON file

    :param data: dict | list Data to be dumped
    :param json_file: str Path to JSON file
    :param indent: int Indent (default: None)
    :param ensure_ascii: bool Whether to encode ASCII symbols

    :return: None
    """

    from json import dump

    with open(json_file, 'w') as file:
        dump(data, file, indent = indent, ensure_ascii = ensure_ascii)


def create_path(path: str) -> None:
    """
    Creates path recursively

    :param path: str Path to directory / file

    :return: None
    """

    from os import makedirs
    from os.path import exists

    # If path does not exist ..
    if not exists(path):
        # .. attempt to ..
        try:
            # .. create it
            makedirs(path)

        # Guard against race condition
        except OSError:
            pass



from typing import Union

from .driver import Driver


class Buzer(Driver):
    """
    Utilities for dealing with 'buzer.de'
    """

    # Individual identifier
    identifier: str = 'buzer'


    def build_title(self, data: dict, mode: Union[bool, str]) -> str:
        """
        Builds description for corresponding legal norm (used as `title` attribute)

        :param data: dict Legal data
        :param mode: str Output mode, either 'light', 'normal' or 'full' (default: False)

        :return: str Title attribute
        :raises: Exception Invalid law
        """

        # Get lowercase identifier for current law
        identifier = data['gesetz'].lower()

        # Fail early if law is unavailable
        if identifier not in self.library:
            raise Exception('Invalid law: "{}"'.format(data['gesetz']))

        # Get data about current law
        law = self.library[identifier]

        # Determine `title` attribute
        if mode == 'light':
            return law['law']

        if mode == 'normal':
            return law['title']

        if mode == 'full':
            return law['headings'][data['norm']['text']]

        return ''


    def build_url(self, data: dict) -> str:
        """
        Builds URL for corresponding legal norm (used as `href` attribute)

        :param data: dict Legal data

        :return: str Target URL
        :raises: Exception Invalid law
        """

        # Get lowercase identifier for current law
        identifier = data['gesetz'].lower()

        # Fail early if law is unavailable
        if identifier not in self.library:
            raise Exception('Invalid law: "{}"'.format(data['gesetz']))

        # Set base URL
        url = 'https://buzer.de'

        # Set HTML file
        file = self.library[identifier]['headings'][data['norm']]['slug']

        # Combine everything
        return '{}/{}'.format(url, file)



from os.path import exists
import re
import time

from .scraper import Scraper
from ..utils import dump_json


class GesetzeImInternet(Scraper):
    """
    Utilities for scraping 'gesetze-im-internet.de'
    """

    # Individual identifier
    identifier: str = 'gesetze'


    def scrape(self, output_file: str, wait: int = 2) -> None:
        """
        Scrapes 'gesetze-im-internet.de' for legal norms

        :param output_file: str Path to merged data file
        :param wait: int Time to wait before scraping next law

        :return: None
        """

        # Create list of data files
        data_files = []

        # Iterate over relevant charset
        for char in 'ABCDEFGHIJKLMNOPQRSTUVWYZ123456789':
            # Fetch overview page for category letter
            html = self.get_html('https://www.gesetze-im-internet.de/Teilliste_{}.html'.format(char))

            # Parse their HTML & iterate over `p` tags ..
            for link in self.bs(html).select('#paddingLR12')[0].select('p'):
                # .. extracting data for each law
                law = link.a.text[1:-1]

                # Define data file for current law
                data_file = self.md5_file(self.dir, law)

                # If it already exists ..
                if exists(data_file):
                    # (1) .. update list of data files
                    data_files.append(data_file)

                    # (2) .. skip law
                    continue

                slug = link.a['href'][2:-11]
                title = link.a.abbr['title']

                # .. collecting its information
                node = {
                    'law': law,
                    'slug': slug,
                    'title': title,
                    'headings': {},
                }

                # Fetch index page for each law
                law_html = self.get_html('https://www.gesetze-im-internet.de/{}/index.html'.format(slug))

                # Iterate over `a` tags ..
                for heading in self.bs(law_html).select('#paddingLR12')[0].select('td'):
                    # (1) .. skipping headings without `a` tag child
                    if not heading.a:
                        continue

                    # (2) .. skipping headings without `href` attribute in `a` tag child
                    if not heading.a.get('href'):
                        continue

                    # Determine section identifier
                    match = re.match(r'(?:§+|Art|Artikel)\.?\s*(\d+(?:\w\b)?)', heading.text, re.IGNORECASE)

                    # If section identifier was found ..
                    if match:
                        # .. store identifier as key and heading as value
                        node['headings'][match.group(1)] = heading.text.strip()

                    # .. otherwise ..
                    else:
                        # .. store heading as both key and value
                        node['headings'][heading.text.strip()] = heading.text.strip()

                # Store data record
                # (1) Write data to JSON file
                dump_json(node, data_file)

                # (2) Update list of data files
                data_files.append(data_file)

                # Wait for it ..
                time.sleep(wait)

        # Merge data files
        self.build(data_files, output_file)



import re
from typing import Callable, Optional, Union

from .factory import Factory
from .helpers import REGEX


class Gesetz:
    """
    Utilities for dealing with german legal norms
    """

    # Available providers
    drivers: dict = {}


    # Defines HTML attribute defaults
    attributes: dict = {'target': '_blank'}


    # Controls `title` attribute
    #
    # Possible values:
    #
    # 'light'  => abbreviated law (eg 'GG')
    # 'normal' => complete law (eg 'Grundgesetz')
    # 'full'   => official heading (eg 'Art 45d Parlamentarisches Kontrollgremium')
    title: Union[bool, str] = False


    def __init__(self, order: Union[list, str] = None) -> None:
        """
        Creates 'Gesetz' instance

        :param order: list | str Single driver OR list of drivers

        :return: None
        """

        # One regEx to rule them all
        self.regex = REGEX

        # Set default order
        if order is None:
            order = [
                'gesetze',     # 'gesetze-im-internet.de'
                'dejure',      # 'dejure.org'
                'buzer',       # 'buzer.de'
                'lexparency',  # 'lexparency.de'
            ]

        # If string was passed as order ..
        if isinstance(order, str):
            # .. make it a list
            order = [order]

        # Initialize drivers
        self.drivers = {driver: Factory().create(driver) for driver in order}


    def validate(self, string: str) -> bool:
        """
        Validates a single legal norm (across all providers)

        :param string: str Legal norm

        :return: bool Whether legal norm is valid (= linkable)
        """

        # Fail early when string is empty
        if not string:
            return False

        # Iterate over drivers
        for driver, obj in self.drivers.items():
            # If legal norm checks out ..
            if obj.validate(string):
                # .. break the loop
                return True

        return False


    def extract(self, string: str) -> list:
        """
        Extracts legal norms as list of strings

        :param string: str Text

        :return: list Extracted legal norms
        """

        return [match[0] for match in self.regex.finditer(string)]


    def linkify(self, match: re.Match) -> str:
        """
        Converts matched legal reference into `a` tag

        :param match: re.Match Matched legal norm

        :return: str Converted `a` tag
        """

        # Import dependencies
        from copy import deepcopy

        # Set `a` tag attribute defaults
        attributes = deepcopy(self.attributes)

        # Fetch extracted data
        string = match.group(0)
        data = match.groupdict()

        # Iterate over drivers for each match ..
        for driver, obj in self.drivers.items():
            # .. using only valid laws & legal norms
            if obj.validate(string):
                # Build `a` tag attributes
                # (1) Determine `href` attribute
                attributes['href'] = obj.build_url(data)

                # (2) Determine `title` attribute
                attributes['title'] = obj.build_title(data, self.title)

                # Abort the loop
                break

        # If `href` attribute is undefined ..
        if 'href' not in attributes:
            # .. return original string
            return string

        # Build `a` tag
        # (1) Format key-value pairs
        attributes = ['{}="{}"'.format(key, value) for key, value in attributes.items() if value]

        # (2) Combine everything
        return '<a {}>{}</a>'.format(' '.join(attributes), string)


    def markdownify(self, match: re.Match) -> str:
        """
        Converts matched legal reference into markdown link

        :param match: re.Match Matched legal norm

        :return: str Converted markdown link
        """

        # Fetch extracted data
        string = match.group(0)
        data = match.groupdict()

        # Set fallback
        link = None

        # Iterate over drivers for each match ..
        for driver, obj in self.drivers.items():
            # .. using only valid laws & legal norms
            if obj.validate(string):
                # Determine link
                link = obj.build_url(data)

                # Abort loop
                break

        # If link is undefined ..
        if not link:
            # .. return original string
            return string

        # Build markdown link
        return '[{}]({})'.format(string, link)


    def gesetzify(self, string: str, callback: Optional[Callable] = None) -> str:
        """
        Converts legal references throughout text into `a` tags

        :param string: str Text
        :param callback: typing.Callable Callback function

        :return: str Processed text
        """

        if callback is None:
            callback = self.linkify

        return self.regex.sub(callback, string)



from .driver import Driver


class GesetzeImInternet(Driver):
    """
    Utilities for dealing with 'gesetze-im-internet.de'
    """

    # Individual identifier
    identifier: str = 'gesetze'


    def build_url(self, data: dict) -> str:
        """
        Builds URL for corresponding legal norm (used as `href` attribute)

        :param data: dict Legal data

        :return: str Target URL
        :raises: Exception Invalid law
        """

        # Get lowercase identifier for current law
        identifier = data['gesetz'].lower()

        # Fail early if law is unavailable
        if identifier not in self.library:
            raise Exception('Invalid law: "{}"'.format(data['gesetz']))

        # Set base URL
        url = 'https://www.gesetze-im-internet.de'

        # Set default HTML file
        file = '__{}.html'.format(data['norm'])

        # Except for the 'Grundgesetz' ..
        if identifier == 'gg':
            # .. which is different
            file = 'art_{}.html'.format(data['norm'])

        # Combine everything
        return '{}/{}/{}'.format(url, self.library[identifier]['slug'], file)



from typing import Optional

import re


# One regEx to rule them all
REGEX = re.compile(r"""
    # Section sign
    (?:§+|&sect;|Art\.?|Artikel)\s*
    # Section ('Norm')
    (?P<norm>\d+(?:\w\b)?)\s*
    # Subsection ('Absatz')
    (?:(?:Abs(?:atz|\.)\s*)?(?P<absatz>(?:\d+|[XIV]+)(?:\w\b)?))?\s*
    # Sentence ('Satz')
    (?:(?:S\.|Satz)\s*(?P<satz>\d+))?\s*
    # Number ('Nummer')
    (?:(?:Nr\.|Nummer)\s*(?P<nr>\d+(?:\w\b)?))?\s*
    # Letter ('Litera')
    (?:(?:lit\.|litera|Buchst\.|Buchstabe)\s*(?P<lit>[a-z]?))?
    # Character limit
    .{0,10}?
    # Law ('Gesetz')
    (?P<gesetz>\b[A-Z][A-Za-z]*[A-Z](?:(?:\s|\b)[XIV]+)?\b)
""", re.VERBOSE)


def analyze(string: str, regex: Optional[re.Pattern] = None) -> dict:
    """
    Analyzes a single legal norm

    :param string: str Legal norm
    :param regex: re.Pattern Regex

    :return: dict Legal norm components
    """

    if regex is None:
        regex = REGEX

    match = regex.search(string)

    if not match:
        return {}

    return {k: v for k, v in match.groupdict().items() if v}


def roman2arabic(string: str) -> int:
    """
    Converts roman numerals to arabic numerals

    :param string: str Roman numeral

    :return: int Arabic numeral
    :raises: Exception Invalid character
    """

    # Transform string to uppercase
    string = string.upper()

    # If one of the characters represents an invalid roman numeral ..
    if not re.match(r'[IVXLCDM]+', string, re.IGNORECASE):
        # .. throw error
        raise Exception('Input contains invalid character!')

    # Map roman base numerals to their arabic equivalent
    romans = {
        'I': 1,
        'V': 5,
        'X': 10,
        'L': 50,
        'C': 100,
        'D': 500,
        'M': 1000,
    }

    # Convert roman numerals to list of arabic numerals
    values = [romans[character] for character in string]

    return sum(
        val if val >= next_val else -val
        for val, next_val in zip(values[:-1], values[1:])
    ) + values[-1]



from .buzer import Buzer
from .dejure import DejureOnline
from .gesetze import GesetzeImInternet
from .lexparency import Lexparency

__all__ = [
    'Buzer',
    'DejureOnline',
    'GesetzeImInternet',
    'Lexparency',
]



import setuptools

# Load README
with open('README.md', 'r', encoding = 'utf8') as file:
    long_description = file.read()

# Define package metadata
setuptools.setup(
    name = 'gesetze',
    version = '1.1.1',
    author = 'Martin Folkers',
    author_email = 'hello@twobrain.io',
    description = '',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://codeberg.org/S1SYPHOS/py-gesetze',
    license = 'MIT',
    project_urls = {
        'Issues': 'https://codeberg.org/S1SYPHOS/py-gesetze/issues',
    },
    entry_points = """
        [console_scripts]
        gesetze=gesetze.cli:cli
    """,
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    packages = setuptools.find_packages(),
    package_data = {'gesetze': ['data/*.json']},
    install_requires = [
        'bs4',
        'click',
        'lxml',
        'requests',
    ],
    python_requires = '>= 3.6'
)



from os.path import exists
import re
import time

from .scraper import Scraper
from ..utils import dump_json


class Lexparency(Scraper):
    """
    Utilities for scraping 'lexparency.de'
    """

    # Individual identifier
    identifier: str = 'lexparency'


    def scrape(self, output_file: str, wait: int = 2) -> None:
        """
        Scrapes 'lexparency.de' for legal norms

        :param output_file: str Path to merged data file
        :param wait: int Time to wait before scraping next law

        :return: None
        """

        # Fetch overview page
        html = self.get_html('https://lexparency.de')

        # Create list of data files
        data_files = []

        # Parse their HTML & iterate over `a` tags ..
        for link in self.bs(html).select('#featured-acts')[0].select('a'):
            # .. extracting data for each law
            law = link.text.strip()

            # Determine abbreviation
            match = re.match(r'.*\((.*)\)$', law)

            # If abbreviation was found ..
            if match:
                # .. store it as shorthand for current law
                law = match.group(1)

            # Define data file for current law
            data_file = self.md5_file(self.dir, law)

            # If it already exists ..
            if exists(data_file):
                # (1) .. update list of data files
                data_files.append(data_file)

                # (2) .. skip law
                continue

            slug = link['href'][4:]

            # .. collecting its information
            node = {
                'law': law,
                'slug': slug,
                'title': '',
                'headings': {},
            }

            # Fetch index page for each law
            law_html = self.get_html('https://lexparency.de/eu/{}'.format(slug))

            # Get title
            # (1) Create empty list
            title = []

            # (2) Convert first character of second entry (= 'title essence') to lowercase
            for i, string in enumerate(list(self.bs(law_html).select('h1')[0].stripped_strings)):
                if i == 1:
                    string = string[:1].lower() + string[1:]

                title.append(string)

            # (3) Create title from strings
            node['title'] = ' '.join(title).strip()

            # Iterate over `li` tags ..
            for heading in self.bs(law_html).select('#toccordion')[0].find_all('li', attrs={'class': 'toc-leaf'}):
                # (1) .. skipping headings without `a` tag child
                if not heading.a:
                    continue

                # (2) .. skipping headings without `href` attribute in `a` tag child
                if not heading.a.get('href'):
                    continue

                string = heading.text.replace('—', '-')

                # Determine section identifier
                match = re.match(r'(?:§+|Art|Artikel)\.?\s*(\d+(?:\w\b)?)', string, re.IGNORECASE)

                # If section identifier was found ..
                if match:
                    # .. store identifier as key and heading as value
                    node['headings'][match.group(1)] = string.replace('§  ', '§ ')

                # .. otherwise ..
                else:
                    # .. store heading as both key and value
                    node['headings'][string] = string

            # Store data record
            # (1) Write data to JSON file
            dump_json(node, data_file)

            # (2) Update list of data files
            data_files.append(data_file)

            # Wait for it ..
            time.sleep(wait)

        # Merge data files
        self.build(data_files, output_file)



from os.path import exists
import re
import time

from .scraper import Scraper
from ..utils import dump_json


class DejureOnline(Scraper):
    """
    Utilities for scraping 'dejure.org'
    """

    # Individual identifier
    identifier: str = 'dejure'


    def scrape(self, output_file: str, wait: int = 2) -> None:
        """
        Scrapes 'dejure.org' for legal norms

        :param output_file: str Path to merged data file
        :param wait: int Time to wait before scraping next law

        :return: None
        """

        # Fetch overview page
        html = self.get_html('https://dejure.org')

        # Create list of data files
        data_files = []

        # Iterate over relevant charset
        for char in 'ABDEFGHIJKLMNOPRSTUVWZ':
            # Parse their HTML & iterate over their sibling's `li` tags ..
            for link in self.bs(html).find('a', attrs={'name': char}).find_next_sibling('ul').select('li'):
                # .. extracting data for each law
                law = link.a.text

                # Define data file for current law
                data_file = self.md5_file(self.dir, law)

                # If it already exists ..
                if exists(data_file):
                    # (1) .. update list of data files
                    data_files.append(data_file)

                    # (2) .. skip law
                    continue

                slug = link.a['href'][9:]
                title = link.text.replace(link.a.text, '').strip('( )')

                # .. collecting its information
                node = {
                    'law': law,
                    'slug': slug,
                    'title': title,
                    'headings': {},
                }

                # Fetch index page for each law
                law_html = self.get_html('https://dejure.org/gesetze/{}'.format(slug))

                # Iterate over `p` tags ..
                for heading in self.bs(law_html).find_all('p', attrs={'class': 'clearfix'}):
                    # (1) .. skipping headings without `a` tag child
                    if not heading.a:
                        continue

                    # (2) .. skipping headings without `href` attribute in `a` tag child
                    if not heading.a.get('href'):
                        continue

                    # Determine section identifier
                    match = re.match(r'(?:§+|Art|Artikel)\.?\s*(\d+(?:\w\b)?)', heading.text, re.IGNORECASE)

                    # If section identifier was found ..
                    if match:
                        # .. store identifier as key and heading as value
                        node['headings'][match.group(1)] = heading.text.strip().replace('§  ', '§ ')

                    # .. otherwise ..
                    else:
                        # .. store heading as both key and value
                        node['headings'][heading.text.strip()] = heading.text.strip()

                # Store data record
                # (1) Write data to JSON file
                dump_json(node, data_file)

                # (2) Update list of data files
                data_files.append(data_file)

                # Wait for it ..
                time.sleep(wait)

            # Waaaait for it ..
            time.sleep(wait * 1.5)

        # Merge data files
        self.build(data_files, output_file)
